#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import formatter
from mock import patch
import six
import test_common

from patsi import ansi
from patsi.document import tree
from patsi.document import loader
from patsi.document import color
from patsi.document import palette
from patsi.document.loader import factory
from patsi.document.loader import _utils


class TestUtils(test_common.TestCase):
    def test_string_to_color(self):
        self.assertEqual(
            _utils.string_to_color("#f0Ba12"),
            color.RgbColor(0xf0, 0xba, 0x12)
        )
        self.assertEqual(
            _utils.string_to_color(""),
            None
        )
        self.assertEqual(
            _utils.string_to_color("blue"),
            palette.colors16.rgb(4)
        )
        self.assertEqual(
            _utils.string_to_color("blue_bright"),
            palette.colors16.rgb(4|8)
        )
        self.assertEqual(
            _utils.string_to_color("NavyBlue"),
            color.IndexedColor(
                palette.colors256.find_index("NavyBlue"),
                palette.colors256,
            )
        )
        self.assertEqual(
            _utils.string_to_color("not_really_a_valid_color"),
            None
        )


class TestFactory(test_common.TestCase):
    loaders = {
        "ansi": loader.AnsiLoader,
        "json": loader.JsonLoader,
        "xml": loader.XmlLoader,
    }

    def test_loader(self):
        for name, cls in six.iteritems(self.loaders):
            self.assertIs(type(factory.loader(name)), cls)
        self.assertRaises(KeyError, lambda: factory.loader("____"))

    def test_loader_for_file(self):
        for name, cls in six.iteritems(self.loaders):
            self.assertIs(type(factory.loader_for_file("foo." + name)), cls)
        self.assertRaises(KeyError, lambda: factory.loader("foo"))

    def test_formats(self):
        self.assertEqual(set(self.loaders.keys()), set(factory.formats()))

    def test_load_file_object(self):
        mock_file = test_common.MockFile()

        self.assertIsInstance(factory.load(mock_file, "ansi"), tree.Document)
        self.assertRaises(KeyError, lambda: factory.load(mock_file, "__"))
        self.assertRaises(Exception, lambda: factory.load(mock_file, "json"))

        del mock_file.name
        self.assertRaises(Exception, lambda: factory.load(mock_file))
        mock_file.name = "foo.ansi"
        self.assertIsInstance(factory.load(mock_file), tree.Document)

    def test_load_file_name(self):
        mock_file = test_common.MockFile()
        with patch("patsi.document.loader.factory.open", mock_file.open):
            self.assertIsInstance(factory.load("foo", "ansi"), tree.Document)
            self.assertRaises(KeyError, lambda: factory.load("foo", "__"))

            self.assertRaises(KeyError, lambda: factory.load("foo"))
            self.assertIsInstance(factory.load("foo.ansi"), tree.Document)


class TestAnsiLoader(test_common.TestCase):
    ldr = loader.AnsiLoader()
    def test_color_from_ansi(self):
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(0)))
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(31, 0)))
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(31, 4)),
            color.IndexedColor(1, palette.colors16)
        )
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(39)))
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.AnsiCode("m", [31])),
            color.IndexedColor(1, palette.colors16)
        )
        self.assertRaises(TypeError, lambda: self.ldr.color_from_ansi(1.2))

        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(31)),
            color.IndexedColor(1, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(91)),
            color.IndexedColor(1|8, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(1, 31)),
            color.IndexedColor(1|8, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(1, 31, 22)),
            color.IndexedColor(1, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(38, 2, 1, 2, 3)),
            color.RgbColor(1, 2, 3)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(38, 5, 42)),
            color.IndexedColor(42, palette.colors256)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(1, 38, 5, 42)),
            color.IndexedColor(42, palette.colors256)
        )

        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(41), True),
            color.IndexedColor(1, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(101), True),
            color.IndexedColor(1|8, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(1, 41), True),
            color.IndexedColor(1|8, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(1, 41, 22), True),
            color.IndexedColor(1, palette.colors16)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(48, 2, 1, 2, 3), True),
            color.RgbColor(1, 2, 3)
        )
        self.assertEqual(
            self.ldr.color_from_ansi(ansi.SGR(48, 5, 42), True),
            color.IndexedColor(42, palette.colors256)
        )

        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(41)))
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(101)))
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(1, 41)))
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(1, 41, 22)))
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(48, 2, 1, 2, 3)))
        self.assertIsNone(self.ldr.color_from_ansi(ansi.SGR(48, 5, 42)))

    def test_load_string(self):
        doc = self.ldr.load_string("Hello")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "Hello\n")

        doc = self.ldr.load_string("Hello\x1b[Hfoo")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "foolo\n")

        doc = self.ldr.load_string("Hello\x1b[31m\nfoo")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 2)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertIsInstance(doc.layers[1], tree.Layer)
        self.assertEqual(doc.layers[0].text, "Hello\n")
        self.assertIsNone(doc.layers[0].color)
        self.assertEqual(doc.layers[1].text, "\nfoo\n")
        self.assertEqual(
            doc.layers[1].color,
            color.IndexedColor(1, palette.colors16)
        )

        doc = self.ldr.load_string("Hello\x1b[38;2;1;2;3m\nfoo")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 2)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertIsInstance(doc.layers[1], tree.Layer)
        self.assertEqual(doc.layers[0].text, "Hello\n")
        self.assertIsNone(doc.layers[0].color)
        self.assertEqual(doc.layers[1].text, "\nfoo\n")
        self.assertEqual(doc.layers[1].color, color.RgbColor(1, 2, 3))

        doc = self.ldr.load_string("Hello\x1b[31m\nfoo\x1b[2J")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 0)

        doc = self.ldr.load_string("Hello\x1b[?25lWorld")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "HelloWorld\n")

    def test_load_string_single_layer(self):
        self.ldr.colors_to_layers = False

        doc = self.ldr.load_string("H\x1b[38;2;1;2;3mell\x1b[31mo")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.FreeColorLayer)
        self.assertEqual(
            doc.layers[0].matrix[(0, 0)],
            ("H", None)
        )
        self.assertEqual(
            doc.layers[0].matrix[(1, 0)],
            ("e", color.RgbColor(1, 2, 3))
        )
        self.assertEqual(
            doc.layers[0].matrix[(2, 0)],
            ("l", color.RgbColor(1, 2, 3))
        )
        self.assertEqual(
            doc.layers[0].matrix[(3, 0)],
            ("l", color.RgbColor(1, 2, 3))
        )
        self.assertEqual(
            doc.layers[0].matrix[(4, 0)],
            ("o", color.IndexedColor(1, palette.colors16))
        )

        self.ldr.colors_to_layers = True

    def test_load_file(self):
        doc = self.ldr.load_file(six.StringIO("Hello"))
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "")
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "Hello\n")

        mock_file = test_common.MockFile(six.StringIO("Hello"))
        with patch("patsi.document.loader.ansi.open", mock_file.open):
            doc = self.ldr.load_file("foo/bar.ansi")
            self.assertIsInstance(doc, tree.Document)
            self.assertEqual(doc.name, "bar")
            self.assertEqual(len(doc.layers), 1)
            self.assertIsInstance(doc.layers[0], tree.Layer)
            self.assertEqual(doc.layers[0].text, "Hello\n")


class TestJsonLoader(test_common.TestCase):
    ldr = loader.JsonLoader()

    def test_document(self):
        doc = self.ldr.load_string("{}")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "")
        self.assertEqual(doc.metadata, {})
        self.assertEqual(len(doc.layers), 0)

        doc = self.ldr.load_string('{"name":"foo"}')
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "foo")
        self.assertEqual(doc.metadata, {})
        self.assertEqual(len(doc.layers), 0)

        doc = self.ldr.load_string('{"name":"foo", "metadata":{"foo":"bar"}}')
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "foo")
        self.assertEqual(doc.metadata, {"foo": "bar"})
        self.assertEqual(len(doc.layers), 0)

    def test_layer(self):
        doc = self.ldr.load_string("""{"layers":[
            {"color": "red", "text": "hello"},
            {"text": "world"},
            {"chars":[
                {"x":0, "y":0, "char":"f"},
                {"x":1, "y":0, "char":"o", "color":"blue"},
                {"x":0, "y":2, "char":"o", "color":"#f00f00"}
            ]}
        ]}""")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 3)

        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "hello\n")
        self.assertEqual(
            doc.layers[0].color,
            color.IndexedColor(1, palette.colors16)
        )

        self.assertIsInstance(doc.layers[1], tree.Layer)
        self.assertEqual(doc.layers[1].text, "world\n")
        self.assertIsNone(doc.layers[1].color)

        self.assertIsInstance(doc.layers[2], tree.FreeColorLayer)
        self.assertEqual(doc.layers[2].matrix[(0, 0)][0], "f")
        self.assertIsNone(doc.layers[2].matrix[(0, 0)][1])
        self.assertEqual(doc.layers[2].matrix[(1, 0)][0], "o")
        self.assertEqual(
            doc.layers[2].matrix[(1, 0)][1],
            color.IndexedColor(4, palette.colors16)
        )
        self.assertEqual(doc.layers[2].matrix[(0, 2)][0], "o")
        self.assertEqual(
            doc.layers[2].matrix[(0, 2)][1],
            color.RgbColor(0xf0, 0x0f, 0x00)
        )

    def test_load_file(self):
        doc_string = '{"layers":[{"text":"Hello"}]}'
        doc = self.ldr.load_file(six.StringIO(doc_string))
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "Hello\n")

        mock_file = test_common.MockFile(six.StringIO(doc_string))
        with patch("patsi.document.loader.json.open", mock_file.open):
            doc = self.ldr.load_file("foo/bar.json")
            self.assertIsInstance(doc, tree.Document)
            self.assertEqual(len(doc.layers), 1)
            self.assertIsInstance(doc.layers[0], tree.Layer)
            self.assertEqual(doc.layers[0].text, "Hello\n")


class TestXmlLoader(test_common.TestCase):
    ldr = loader.XmlLoader()

    def test_document(self):
        doc = self.ldr.load_string("<document></document>")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "")
        self.assertEqual(doc.metadata, {})
        self.assertEqual(len(doc.layers), 0)

        doc = self.ldr.load_string("""<document name="foo"></document>""")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "foo")
        self.assertEqual(doc.metadata, {})
        self.assertEqual(len(doc.layers), 0)

        doc = self.ldr.load_string("""<document name="foo">
                <metadata><foo>bar</foo></metadata>
            </document>""")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(doc.name, "foo")
        self.assertEqual(doc.metadata, {"foo": "bar"})
        self.assertEqual(len(doc.layers), 0)

    def test_layer(self):
        doc = self.ldr.load_string("""<document>
            <layer color="red">hello</layer>
            <layer>world</layer>
            <layer>
                <char x="0" y="0">f</char>
                <char x="1" y="0" color="blue">o</char>
                <char x="0" y="2" color="#f00f00">o</char>
            </layer>
        </document>""")
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 3)

        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "hello\n")
        self.assertEqual(
            doc.layers[0].color,
            color.IndexedColor(1, palette.colors16)
        )

        self.assertIsInstance(doc.layers[1], tree.Layer)
        self.assertEqual(doc.layers[1].text, "world\n")
        self.assertIsNone(doc.layers[1].color)

        self.assertIsInstance(doc.layers[2], tree.FreeColorLayer)
        self.assertEqual(doc.layers[2].matrix[(0, 0)][0], "f")
        self.assertIsNone(doc.layers[2].matrix[(0, 0)][1])
        self.assertEqual(doc.layers[2].matrix[(1, 0)][0], "o")
        self.assertEqual(
            doc.layers[2].matrix[(1, 0)][1],
            color.IndexedColor(4, palette.colors16)
        )
        self.assertEqual(doc.layers[2].matrix[(0, 2)][0], "o")
        self.assertEqual(
            doc.layers[2].matrix[(0, 2)][1],
            color.RgbColor(0xf0, 0x0f, 0x00)
        )

    def test_load_file(self):
        doc_string = '<document><layer>Hello</layer></document>'
        doc = self.ldr.load_file(six.StringIO(doc_string))
        self.assertIsInstance(doc, tree.Document)
        self.assertEqual(len(doc.layers), 1)
        self.assertIsInstance(doc.layers[0], tree.Layer)
        self.assertEqual(doc.layers[0].text, "Hello\n")


test_common.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import sys
import tty
import termios
import contextlib


class RawInput(object):
    def __init__(self, read_ansi=True, raise_specials=True):
        self.file = sys.stdin
        self.fileno = self.file.fileno()
        self.attrs = None
        self.when = termios.TCSAFLUSH
        self.read_ansi = True
        self.raise_specials = True

    def get(self):
        ch = self.file.read(1)
        code = ord(ch)
        if code == 4 and self.raise_specials:
            raise EOFError()
        if  code == 3 and self.raise_specials:
            raise KeyboardInterrupt()
        if code == 0x1b and self.read_ansi:
            with self.nonblocking():
                ch += self.file.readline()
        return ch

    def __iter__(self):
        while True:
            yield self.get()

    def __enter__(self):
        if self.attrs is None:
            self.attrs = termios.tcgetattr(self.fileno)
            tty.setraw(self.fileno, self.when)
        return self

    def __exit__(self, *a, **k):
        if self.attrs:
            termios.tcsetattr(self.fileno, self.when, self.attrs)

    @contextlib.contextmanager
    def nonblocking(self):
        mode = termios.tcgetattr(self.fileno)
        try:
            mode[tty.CC][termios.VMIN] = 0
            termios.tcsetattr(sys.stdin, self.when, mode)
            yield
        finally:
            mode[tty.CC][termios.VMIN] = 1
            termios.tcsetattr(sys.stdin, self.when, mode)



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import test_common

from patsi.document.tree import *
from patsi.document import color
from patsi.document import _misc


class TestLayer(test_common.TestCase):
    def test_empty(self):
        l = Layer()
        self.assertEqual(l.text, "")
        self.assertEqual(l.color, None)
        self.assertEqual(l.width, 0)
        self.assertEqual(l.height, 0)

    def test_set_char(self):
        l = Layer("foo\nBar!\n")
        self.assertEqual(l.text, "foo\nBar!\n")
        self.assertEqual(l.color, None)
        self.assertEqual(l.width, 4)
        self.assertEqual(l.height, 2)
        self.assertEqual(l.lines, ["foo", "Bar!"])
        self.assertEqual(l.lines[1][3], "!")
        l.set_char(1, 0, "X")
        self.assertEqual(l.lines[0][1], "X")
        self.assertEqual(l.width, 4)
        self.assertEqual(l.height, 2)
        l.set_char(5, 5, "X")
        self.assertEqual(l.lines[5][5], "X")
        self.assertEqual(l.width, 6)
        self.assertEqual(l.height, 6)

    def test_empty_trailing_line(self):
        l = Layer("foo\nBar!\n\n")
        self.assertEqual(l.text, "foo\nBar!\n\n")
        self.assertEqual(l.height, 3)
        self.assertEqual(l.lines, ["foo", "Bar!", ""])

    def test_no_trailing_newline(self):
        l = Layer("foo\nBar!")
        self.assertEqual(l.text, "foo\nBar!\n")
        self.assertEqual(l.color, None)
        self.assertEqual(l.width, 4)
        self.assertEqual(l.height, 2)
        self.assertEqual(l.lines, ["foo", "Bar!"])

    def test_color(self):
        l = Layer("foo\nBar!\n", color.RgbColor(1, 2, 3))
        self.assertEqual(l.text, "foo\nBar!\n")
        self.assertEqual(l.color, color.RgbColor(1, 2, 3))
        self.assertEqual(l.width, 4)
        self.assertEqual(l.height, 2)


class TestFreeColorLayer(test_common.TestCase):
    def test_empty(self):
        l = FreeColorLayer()
        self.assertFalse(l.matrix)
        self.assertEqual(l.width, 0)
        self.assertEqual(l.height, 0)

    def test_set_char(self):
        l = FreeColorLayer()
        self.assertFalse((4, 5) in l.matrix)
        l.set_char(4, 5, "x", color.RgbColor(1, 2, 3))
        self.assertTrue((4, 5) in l.matrix)
        self.assertEqual(l.matrix[(4, 5)], ("x", color.RgbColor(1, 2, 3)))
        self.assertEqual(l.width, 5)
        self.assertEqual(l.height, 6)

        l.set_char(4, 5, "y", color.RgbColor(3, 2, 1))
        self.assertEqual(l.matrix[(4, 5)], ("y", color.RgbColor(3, 2, 1)))
        self.assertEqual(l.width, 5)
        self.assertEqual(l.height, 6)

        l.set_char(1, 1, "y")
        self.assertEqual(l.matrix[(1, 1)], ("y", None))
        self.assertEqual(l.width, 5)
        self.assertEqual(l.height, 6)

    def test_add_layer(self):
        l = FreeColorLayer()
        color_foo = color.RgbColor(0xf, 0, 0)
        l.add_layer(Layer("Foo", color_foo))
        self.assertEqual(l.matrix[(0, 0)], ("F", color_foo))
        self.assertEqual(l.matrix[(1, 0)], ("o", color_foo))
        self.assertEqual(l.matrix[(2, 0)], ("o", color_foo))

        color_ubar = color.RgbColor(0xb, 0xa, 0x2)
        l.add_layer(Layer(" u\n b\n a\n r", color_ubar))
        self.assertEqual(l.matrix[(0, 0)], ("F", color_foo))
        self.assertEqual(l.matrix[(1, 0)], ("u", color_ubar))
        self.assertEqual(l.matrix[(2, 0)], ("o", color_foo))
        self.assertEqual(l.matrix[(1, 1)], ("b", color_ubar))

        color_fun = color.RgbColor(0xf, 0xa, 0xf)
        l.add_layer(Layer("  n\n", color_fun))
        self.assertEqual(l.matrix[(0, 0)], ("F", color_foo))
        self.assertEqual(l.matrix[(1, 0)], ("u", color_ubar))
        self.assertEqual(l.matrix[(2, 0)], ("n", color_fun))

        other = FreeColorLayer()
        other.set_char(3, 0, "!", color_foo)
        l.add_layer(other)
        self.assertEqual(l.matrix[(0, 0)], ("F", color_foo))
        self.assertEqual(l.matrix[(1, 0)], ("u", color_ubar))
        self.assertEqual(l.matrix[(2, 0)], ("n", color_fun))
        self.assertEqual(l.matrix[(3, 0)], ("!", color_foo))

        self.assertRaises(TypeError, lambda: l.add_layer(None))


class TestDocument(test_common.TestCase):
    def test_ctor(self):
        doc = Document()
        self.assertEqual(doc.name, "")
        self.assertEqual(doc.layers, [])
        self.assertEqual(doc.metadata, {})

        doc = Document("foo")
        self.assertEqual(doc.name, "foo")
        self.assertEqual(doc.layers, [])
        self.assertEqual(doc.metadata, {})

        doc = Document("foo", [Layer()], {"foo": "bar"})
        self.assertEqual(doc.name, "foo")
        self.assertEqual(len(doc.layers), 1)
        self.assertEqual(doc.metadata, {"foo": "bar"})

    def test_flatten(self):
        doc = Document()
        layer = doc.flattened()
        self.assertIsInstance(layer, Layer)
        self.assertEqual(layer.text, "")

        doc.layers.append(Layer("Foo\n", color.RgbColor(1, 2, 3)))
        layer = doc.flattened()
        self.assertIsInstance(layer, Layer)
        self.assertEqual(layer.text, "Foo\n")
        self.assertEqual(layer.color, doc.layers[0].color)

        doc.layers.append(Layer("\nbar\n", color.RgbColor(3, 2, 1)))
        layer = doc.flattened()
        self.assertIsInstance(layer, FreeColorLayer)
        self.assertEqual(layer.matrix[(0, 0)], ("F", color.RgbColor(1, 2, 3)))
        self.assertEqual(layer.matrix[(0, 1)], ("b", color.RgbColor(3, 2, 1)))

        doc2 = doc.flattened_doc()
        self.assertEqual(len(doc2.layers), 1)
        self.assertIsInstance(doc2.layers[0], FreeColorLayer)
        self.assertEqual(doc2.layers[0].matrix, layer.matrix)

    def test_properties(self):
        doc = Document(layers=[Layer("Foo\n\nBar"), Layer("Hello")])
        self.assertEqual(doc.width, 5)
        self.assertEqual(doc.height, 3)


class TestMisc(test_common.TestCase):
    def test_basename(self):
        self.assertEqual(_misc.basename(""), "")
        self.assertEqual(_misc.basename("foo"), "foo")
        self.assertEqual(_misc.basename("foo.bar"), "foo")
        self.assertEqual(_misc.basename("foo.b.a.r"), "foo")
        self.assertEqual(_misc.basename("hello/foo.bar"), "foo")
        self.assertEqual(_misc.basename("hello/foo"), "foo")
        self.assertEqual(_misc.basename("hello.world/foo"), "foo")
        self.assertEqual(_misc.basename("hello.world/foo.bar"), "foo")

    def test_extension(self):
        self.assertEqual(_misc.extension("foo"), "")
        self.assertEqual(_misc.extension("foo.bar"), "bar")
        self.assertEqual(_misc.extension("hello/foo.bar"), "bar")
        self.assertEqual(_misc.extension("hello/foo.b.a.r"), "b.a.r")
        self.assertEqual(_misc.extension("hello.world/foo.bar"), "bar")


test_common.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from . import factory
from .. import tree
from .. import _misc
from .. import color as color_ns
from .. import palette
from ... import ansi


class AnsiLoader(object):
    class _LayerTracker(object):
        def __init__(self, loader, doc_name):
            self.doc = tree.Document(doc_name)
            self.loader = loader
            self.clear()

        def clear(self):
            if self.loader.colors_to_layers:
                self.doc.layers = []
                self.layers = {}
            else:
                self.doc.layers = [tree.FreeColorLayer()]

        def set_char(self, x, y, ch, color):
            if not self.loader.colors_to_layers:
                self.doc.layers[0].set_char(x, y, ch, color)
                return

            if color is None:
                index = (-1, None)
            elif isinstance(color, color_ns.IndexedColor):
                index = (len(color.palette), color.index)
            else:
                index = (-1, color.rgb)

            if index not in self.layers:
                layer = tree.Layer("", color)
                self.layers[index] = layer
                self.doc.layers.append(layer)

            self.layers[index].set_char(x, y, ch)

    def __init__(self, colors_to_layers=True):
        self.colors_to_layers = colors_to_layers

    def load_string(self, string, doc_name=""):
        layer_tracker = AnsiLoader._LayerTracker(self, doc_name)

        color = None
        mover = ansi.CharMover(0, 0)
        for item in ansi.AnsiCode.split(string):
            if isinstance(item, ansi.CursorPosition):
                mover.move(item.column - 1, item.row - 1)
            elif isinstance(item, ansi.SGR):
                color = AnsiLoader.color_from_ansi(item)
            elif isinstance(item, ansi.AnsiCode):
                if item == ansi.common.clear_screen:
                    layer_tracker.clear()
            else:
                for ch in mover.loop(item):
                    layer_tracker.set_char(mover.x, mover.y, ch, color)
        return layer_tracker.doc

    def load_file(self, file):
        if type(file) is str:
            with open(file) as f:
                return self.load_file(f)
        else:
            if hasattr(file, "name"):
                name = _misc.basename(file.name)
            else:
                name = ""
            return self.load_string(file.read(), name)

    @staticmethod
    def color_from_ansi(color, background=False):
        if not isinstance(color, ansi.SGR):
            if isinstance(color, ansi.AnsiCode) and color.terminator == 'm':
                color = ansi.SGR(int(arg) for arg in color.args())
            else:
                raise TypeError("Expected SGR ansi code object")

        result = None
        bright = False
        for flag in color.flags:
            if flag == 0:
                result = None
                bright = False
            elif flag == 1:
                bright = True
            elif flag == 22:
                bright = False
            elif isinstance(flag, ansi.SGR.ColorRGB):
                if flag.background == background:
                    result = color_ns.RgbColor(flag.r, flag.g, flag.b)
            elif isinstance(flag, ansi.SGR.Color):
                if flag.background == background:
                    if flag.color == flag.ResetColor:
                        result = None
                    else:
                        bright = flag.bright or bright
                        result = color_ns.IndexedColor(flag.color, palette.colors16)
            elif isinstance(flag, ansi.SGR.Color256):
                if flag.background == background:
                    result = color_ns.IndexedColor(flag.color, palette.colors256)

        if isinstance(result, color_ns.IndexedColor) and len(result.palette) == 16 and bright:
            result.index += 8

        return result


factory.register(AnsiLoader(), "ansi")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
import json
from . import factory
from .. import tree


class JsonFormatter(object):
    flat = False

    def __init__(self, indent=4, separators=None, sort_keys=False):
        self.indent = indent
        self.separators = separators
        self.sort_keys = sort_keys

    def document_dict(self, doc):
        return {
            "name": doc.name,
            "metadata": doc.metadata,
            "layers": [self.layer_dict(layer) for layer in doc.layers]
        }

    def layer_dict(self, layer):
        if isinstance(layer, tree.Layer):
            return {
                "color": self.color_dict(layer.color),
                "text": layer.text
            }
        elif isinstance(layer, tree.FreeColorLayer):
            return {
                "chars": [
                    {
                        "x": pos[0],
                        "y": pos[1],
                        "char": layer.matrix[pos][0],
                        "color": self.color_dict(layer.matrix[pos][1]),
                    }
                    for pos in
                    sorted(layer.matrix.keys(), key=lambda a: (a[1], a[0]))
                ]
            }
        else:
            raise TypeError("Expected layer type")

    def color_dict(self, color):
        if color is None:
            return ""
        return color.name

    def _json_args(self):
        return {
            "indent": self.indent,
            "separators": self.separators,
            "sort_keys": self.sort_keys,
        }

    def document(self, doc, output):
        json.dump(self.document_dict(doc), output, **self._json_args())

    def layer(self, layer, output):
        json.dump(self.layer_dict(layer), output, **self._json_args())

    def color(self, color):
        return json.dumps(self.color_dict(color), **self._json_args())


factory.register(JsonFormatter(), "json")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os.path


def _split(file_name):
    return os.path.basename(file_name).split(".", 1)


def basename(file_name):
    return _split(file_name)[0]


def extension(file_name):
    split = _split(file_name)
    return split[1] if len(split) > 1 else ""



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from .. import tree
from .. import _misc


_formatters = {}


def register(formatter, file_extension):
    _formatters[file_extension] = formatter


def formatter(file_extension):
    return _formatters[file_extension]


def formatter_for_file(file_name):
    return formatter(_misc.extension(file_name))


def save(document, file, hint=None):
    if type(file) is str:
        with open(file, "w") as file_obj:
            save(document, file_obj, hint)
            return

    if hint is not None:
        fmt = formatter(hint)
    else:
        fmt = formatter_for_file(file.name)

    if isinstance(document, tree.Document):
        fmt.document(document, file)
    else:
        fmt.layer(document, file)


def formats():
    return sorted(_formatters.keys())



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
from ..ansi import CharMover


class Document(object):
    def __init__(self, name="", layers=[], metadata={}):
        self.name = name
        self.layers = layers if layers else []       # This avoids overwriting the default argument value
        self.metadata = metadata if metadata else {}

    def flattened(self):
        if len(self.layers) == 0:
            return Layer()
        if len(self.layers) == 1:
            return copy.deepcopy(self.layers[0])

        flattened = FreeColorLayer(
            max(layer.width for layer in self.layers),
            max(layer.height for layer in self.layers)
        )

        for layer in self.layers:
            flattened.add_layer(layer)

        return flattened

    def flattened_doc(self):
        return Document(self.name, [self.flattened()])

    @property
    def width(self):
        return max(layer.width for layer in self.layers)

    @property
    def height(self):
        return max(layer.height for layer in self.layers)


class Layer(object):
    def __init__(self, text="", color=None):
        if text.endswith("\n"):
            text = text[:-1]
        self.lines = text.split("\n") if text else []
        self.color = color

    @property
    def width(self):
        return max(len(line) for line in self.lines) if self.lines else 0

    @property
    def height(self):
        return len(self.lines)

    @property
    def text(self):
        if not self.lines:
            return ""
        return "\n".join(self.lines) + "\n"

    def set_char(self, x, y, char):
        if y >= self.height:
            self.lines += [""] * (y - self.height + 1)
        if x >= len(self.lines[y]):
            self.lines[y] += " " * (x - len(self.lines[y]) + 1)

        self.lines[y] = self.lines[y][:x] + char + self.lines[y][x + 1:]


class FreeColorLayer(object):
    def __init__(self, width=0, height=0):
        self.matrix = {}
        self.width = width
        self.height = height

    def set_char(self, x, y, char, color=None):
        if y >= self.height:
            self.height = y + 1
        if x >= self.width:
            self.width = x + 1

        self.matrix[(x, y)] = (char, color)

    def add_layer(self, layer, offset_x=0, offset_y=0):
        mover = CharMover(offset_x, offset_y)
        if isinstance(layer, Layer):
            for ch in mover.loop(layer.text):
                self.set_char(mover.x, mover.y, ch, layer.color)
        elif isinstance(layer, FreeColorLayer):
            self.matrix.update(layer.matrix)
        else:
            raise TypeError("Expected Layer")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
import xml.etree.cElementTree as ElementTree

from . import factory
from ._utils import string_to_color
from .. import tree


class XmlLoader(object):
    def _load_xml_layer(self, element):
        if len(element):
            layer = tree.FreeColorLayer()
            for char in element.findall("char"):
                layer.set_char(
                    int(char.attrib["x"]),
                    int(char.attrib["y"]),
                    char.text,
                    string_to_color(char.get("color", ""))
                )
            return layer

        return tree.Layer(
            element.text,
            string_to_color(element.get("color", ""))
        )

    def _load_xml_metadata(self, element):
        if element is not None:
            return {child.tag: child.text for child in element}
        return {}

    def load_xml(self, doc_element):
        return tree.Document(
            doc_element.get("name", ""),
            [self._load_xml_layer(layer) for layer in doc_element.findall("layer")],
            self._load_xml_metadata(doc_element.find("metadata")),
        )

    def load_string(self, string):
        return self.load_xml(ElementTree.fromstring(string))

    def load_file(self, file):
        return self.load_xml(ElementTree.parse(file).getroot())


factory.register(XmlLoader(), "xml")



#!/usr/bin/env python
#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import argparse
from patsi.document import tree, formatter, loader, _misc, color, palette


def load_dir(path):
    doc = tree.Document(_misc.basename(path))
    for file in os.listdir(path):
        if file in palette.colors16.names:
            doc.layers.append(tree.Layer(
                open(os.path.join(path, file)).read(),
                color.IndexedColor(file, palette.colors16)
            ))
    return doc


def parse_args():
    parser = argparse.ArgumentParser(
        description="Renders an patsi document."
    )

    parser.add_argument(
        "--debug",
        help="Enable debugging",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--input", "-i",
        type=str,
        help="Input file (Use - for standard output)",
        default="-",
    )

    parser.add_argument(
        "--input-format", "-if",
        help="Input format",
        default="auto",
        choices=["auto", "ansi_dir"] + loader.factory.formats(),
    )

    # auto makes sense when -o is defined
    parser.add_argument(
        "--output-format", "-of",
        help="Output format",
        default="auto",
        choices=["auto"] + formatter.factory.formats(),
    )

    parser.add_argument(
        "--ansi",
        help="Ansi output features",
        default=[],
        #choices=formatter.AnsiFormatter.Features.constants_dict().keys(),
        nargs="+",
        dest="ansi_features",
    )

    parser_output = parser.add_mutually_exclusive_group()

    # If -of is defined, default to output_directory/basename(input).output_format
    parser_output.add_argument(
        "--output", "-o",
        help="Output file (Use - for standard output)",
        default="",
    )

    # Defaults to dirname(input)
    parser_output.add_argument(
        "--output-directory", "-d",
        help="Output directory (Used when -o is omitted)",
        default=None,
    )

    return parser.parse_args()


class RenderException(Exception):
    pass


def run(args):
    uses_stdin = args.input == "-" or not args.input

    if args.ansi_features:
        formatter.AnsiFormatter.features = formatter.AnsiFormatter.Features(args.ansi_features)

    # Load the document
    if uses_stdin:
        if args.input_format == "ansi_dir":
            raise RenderException("You must pass a path to open an ANSI directory")
        elif args.input_format == "auto":
            # TODO Avoid this if the factory is set to use libmagic
            raise RenderException("Cannot autodetect format from standard input")
        doc = loader.factory.load(sys.stdin, args.input_format)
    else:
        if args.input_format == "ansi_dir" or \
                (args.input_format == "auto" and os.path.isdir(args.input)):
            doc = load_dir(args.input)
        else:
            hint = None if args.input_format == "auto" else args.input_format
            doc = loader.factory.load(args.input, hint)

    # Get the file name
    if not args.output:
        if args.output_format == "auto":
            raise RenderException("You must give an explicit file name to autodetect the output format")

        if args.output_directory is not None:
            dirname = args.output_directory
        elif uses_stdin:
            dirname = os.getcwd()
        else:
            dirname = os.path.dirname(args.input)

        if uses_stdin:
            basename = "stdin" if not doc.name else doc.name
        else:
            basename = _misc.basename(args.input)

        output_name = os.path.join(dirname, basename) + "." + args.output_format
    else:
        output_name = args.output

    if output_name == "-" and args.output_format == "auto":
        args.output_format = "ansi"

    # Write to output
    out = sys.stdout if output_name == "-" else output_name
    hint = None if args.output_format == "auto" else args.output_format
    formatter.factory.save(doc, out, hint)


def main():
    args = parse_args()
    with ExceptionManager(args.debug):
        run(args)


class ExceptionManager(object):
    def __init__(self, debug=False):
        self.debug = debug

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc, traceback):
        if self.debug or exc is None:
            return
        if isinstance(exc, RenderException) or isinstance(exc, IOError):
            self._error(exc)
        elif isinstance(exc, KeyError):
            self._error("Unrecognized format: %s" % exc)
        else:
            self._error("Unexpected error")

    def _error(self, msg):
        sys.stderr.write(str(msg) + "\n")
        sys.exit(1)


main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import re
from .. import color as color_ns
from .. import palette


_rgb_color_regex = re.compile("^#[0-9a-fA-F]{6}$")


def string_to_color(color):
    if not color:
        return None
    if _rgb_color_regex.match(color):
        return color_ns.RgbColor(
            int(color[1:3], 16),
            int(color[3:5], 16),
            int(color[5:7], 16)
        )
    if color in palette.colors16.names:
        return color_ns.IndexedColor(str(color), palette.colors16)
    if color in palette.colors256.names:
        return color_ns.IndexedColor(str(color), palette.colors256)
    return None



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from .widget import Widget
from .geometry import Point, Rect
import curses
from ._util import next_object


class TabBar(Widget):
    def __init__(self, parent, window, items, current, formatter=str, separator="|"):
        super(TabBar, self).__init__(parent, window)
        self.items = items
        self.current = current
        self.formatter = formatter
        self.separator = separator
        self.signal_changed = lambda x: None
        self.render()

    def render(self):
        p = Point()
        self._draw_string(p, self.separator)
        p.x += len(self.separator)
        for item in self.items:
            if item is not self.current:
                mode = curses.A_NORMAL
            elif self.has_focus():
                mode = curses.A_REVERSE
            else:
                mode = curses.A_UNDERLINE
            text = self.formatter(item)
            self._draw_string(p, text, mode)
            p.x += len(text)
            self._draw_string(p, self.separator)
            p.x += len(self.separator)
        # TODO handle overflow

    def key_event(self, event):
        if event.key == curses.KEY_LEFT:
            self._switch_element(-1)
        elif event.key == curses.KEY_RIGHT:
            self._switch_element(+1)
        elif event.key == curses.KEY_HOME:
            if not self.items:
                self._change_current(None)
            else:
                self._change_current(self.items[0])
        elif event.key == curses.KEY_END:
            if not self.items:
                self._change_current(None)
            else:
                self._change_current(self.items[-1])

    def _switch_element(self, delta):
        if not self.items:
            self._change_current(None)
        else:
            self._change_current(next_object(self.items, self.current, delta))

    def window_bounds_hint(self, parent_bounds):
        return Rect(
            top_left=parent_bounds.top_left,
            bottom_right=Point(
                parent_bounds.right,
                parent_bounds.top + 1,
            )
        )

    def _change_current(self, next):
        if next != self.current:
            self.current = next
            self.signal_changed(self.current)
            self.refresh()

    def mouse_event(self, event):
        super(TabBar, self).mouse_event(event)
        if event.buttons == curses.BUTTON1_CLICKED:
            x = len(self.separator)
            for item in self.items:
                length = len(self.formatter(item))
                if x < event.pos.x < x + length:
                    self._change_current(item)
                    break
                x += length
            event.accept()


class TabArea(Widget):
    def window_bounds_hint(self, parent_bounds):
        bounds = parent_bounds.copy()
        bounds.y1 = bounds.y1 + 1
        return bounds


class TabWidget(Widget):
    def __init__(self, parent, window, formatter=lambda x: "tab"):
        super(TabWidget, self).__init__(parent, window)

        self.current_tab = None
        self.container = TabArea(self)
        self.tabs = self.container.children

        self.bar = TabBar(self, None, self.tabs, None, formatter)
        self.bar.signal_changed = self._switch_current
        self.bar.items = self.tabs
        self.focus(self.bar)

    def close_current(self):
        if not self.current_tab:
            return

        index = self.tabs.index(self.current_tab)
        self.tabs.remove(self.current_tab)
        self.current_tab = None
        if self.tabs:
            if index == len(self.tabs):
                index -= 1
            self._switch_current(self.tabs[index])
        else:
            self.refresh()

    def create_tab(self, widget_class=Widget, *args, **kwargs):
        widget = self.create_tab_noadd(widget_class, *args, **kwargs)
        self.add_tab(widget)
        return widget

    def create_tab_noadd(self, widget_class=Widget, *args, **kwargs):
        return widget_class(None, self.container.window, *args, **kwargs)

    def add_tab(self, widget):
        widget.parent = self.container
        if widget not in self.tabs:
            self.tabs.append(widget)
        self._switch_current(widget)
        self.activate_tab()

    def _switch_current(self, tab):
        if tab != self.current_tab:

            old_tab = self.current_tab
            if old_tab:
                old_tab.active = False

            self.current_tab = tab
            self.bar.current = tab

            if tab:
                tab.active = True
                self.container.focus(tab)

            self.refresh()

    def on_activated(self, tab):
        pass

    def activate_tab(self):
        if self.current_tab:
            self.current_tab.active = True
            self.focus(self.container)
            self.on_activated(self.current_tab)
            self.current_tab.refresh()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
import six
from xml.sax.saxutils import escape
from contextlib import contextmanager

from . import factory
from .. import tree
from .. import color
from .. import palette


class SvgFormatter(object):

    def __init__(
            self,
            flatten=False,
            background=color.IndexedColor(0, palette.colors8_dark),
            text_color=color.IndexedColor(7, palette.colors8_dark),
            font_size=12):
        self.flat = flatten
        self.background = background
        self.text_color = text_color
        self.font_size = font_size

    @property
    def font_width(self):
        return self.font_size / 2.0

    @property
    def font_height(self):
        return self.font_size

    def document(self, doc, output):
        with self.wrap_svg(doc, output):
            if self.flat:
                self.layer(doc.flattened(), output, True)
            else:
                for layer in doc.layers:
                    self.layer(layer, output, True)

    @contextmanager
    def wrap_svg(self, element, output):
        width = element.width * self.font_width
        height = element.height * self.font_height

        output.write("<?xml version='1.0' encoding='UTF-8' ?>\n")
        output.write("<svg xmlns='http://www.w3.org/2000/svg' width='%s' height='%s'>\n" % (width, height))
        output.write("<rect style='fill:%s;stroke:none;' width='%s' height='%s' x='0' y='0' />\n" % (
            self.color(self.background),
            width,
            height
        ))

        yield

        output.write("</svg>\n")

    def layer(self, layer, output, bare=False):
        if not bare and (
                isinstance(layer, tree.Layer) or
                isinstance(layer, tree.FreeColorLayer)
        ):
            with self.wrap_svg(layer, output):
                return self.layer(layer, output, True)

        css = [
            "font-family:monospace",
            "font-size:%spx" % self.font_size,
            "font-weight:%s" % self.color(self.text_color),
            "fill:white",
        ]

        def open_rect():
            return "<text y='0' x='0' style='%s' xml:space='preserve'>\n" % ";".join(css)

        if isinstance(layer, tree.Layer):
            css.append("fill:%s" % self.color(layer.color))
            css.append("letter-spacing:-%spx" % (self.font_width * 0.2))
            output.write(open_rect())
            y = 0
            for line in layer.lines:
                y += 1
                if line:
                    output.write("<tspan x='{x}' y='{y}'>{line}</tspan>\n".format(
                        x=0,
                        y=y * self.font_height,
                        line=escape(line)
                    ))
        elif isinstance(layer, tree.FreeColorLayer):
            output.write(open_rect())
            prev_color = None
            for pos, item in six.iteritems(layer.matrix):
                char, col = item
                if col is not color.UnchangedColor and col != prev_color:
                    prev_color = col
                output.write("<tspan x='{x}' y='{y}' style='fill:{color};'>{char}</tspan>\n".format(
                    x=pos[0] * self.font_width,
                    y=pos[1] * self.font_height,
                    color=self.color(prev_color),
                    char=escape(char)
                ))
        else:
            raise TypeError("Expected layer type")

        output.write("</text>\n")

    def color(self, color):
        if color is None:
            return "inherit"
        return color.rgb.hex()


factory.register(SvgFormatter(), "svg")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import test_common

from patsi.document import palette
from patsi.document.color import *


class TestPalette(test_common.TestCase):
    names = ["foo", "bar"]
    colors = [(0xf, 0, 0), (0xb, 0xa, 2)]
    zipped = list(zip(names, colors))

    def test_ctor(self):
        p = palette.Palette(self.zipped)
        self.assertEqual(p.names, self.names)
        self.assertEqual(p.colors, self.colors)

        p = palette.Palette(*self.zipped)
        self.assertEqual(p.names, self.names)
        self.assertEqual(p.colors, self.colors)

        p = palette.Palette(item for item in self.zipped)
        self.assertEqual(p.names, self.names)
        self.assertEqual(p.colors, self.colors)

    def test_rgb(self):
        p = palette.Palette(self.zipped)
        self.assertEqual(p.rgb(0), self.colors[0])
        self.assertEqual(p.rgb(1), self.colors[1])
        self.assertRaises(IndexError, p.rgb, len(self.colors))

    def test_name(self):
        p = palette.Palette(self.zipped)
        self.assertEqual(p.name(0), self.names[0])
        self.assertEqual(p.name(1), self.names[1])
        self.assertRaises(IndexError, p.rgb, len(self.names))

    def test_iter(self):
        self.assertEqual(list(palette.Palette(self.zipped)), self.zipped)

    def test_add(self):
        other_names = ["fu", "ba", "foo"]
        other_colors = [(1,2,3), (3,4,5), (4,5,6)]
        other = palette.Palette(zip(other_names, other_colors))

        p = palette.Palette(self.zipped)
        p += other
        self.assertEqual(p.names, ["foo", "bar", "fu", "ba"])
        self.assertEqual(p.colors, [(4,5,6), (0xb, 0xa, 2), (1,2,3), (3,4,5)])

        q = palette.Palette(self.zipped)
        r = q + other
        self.assertEqual(list(q), self.zipped)
        self.assertEqual(list(r), list(p))

    def test_len(self):
        p = palette.Palette(self.zipped)
        self.assertEqual(len(p), len(self.colors))

    def test_builtin(self):
        self.assertEqual(len(palette.colors8_dark), 8)
        self.assertEqual(len(palette.colors8_bright), 8)
        self.assertEqual(len(palette.colors16), 16)
        self.assertEqual(len(palette.colors256), 256)

    def test_find_index(self):
        p = palette.Palette(self.zipped)
        self.assertEqual(p.find_index("bar"), 1)
        self.assertEqual(p.find_index(self.colors[1]), 1)

        self.assertRaises(Exception, p.find_index, "barr")


class TestIndexedColor(test_common.TestCase):
    def test_ctor(self):
        c = IndexedColor(2, palette.colors8_dark)
        self.assertEqual(c.index, 2)
        self.assertIs(c.palette, palette.colors8_dark)

        c = IndexedColor("blue", palette.colors8_dark)
        self.assertEqual(c.index, 4)
        self.assertIs(c.palette, palette.colors8_dark)

    def test_rgb(self):
        color = IndexedColor(2, palette.colors8_dark)
        self.assertIs(type(color.rgb), RgbColor)
        self.assertEqual(color.rgb, palette.colors8_dark.rgb(2))

    def test_rgb_tuple(self):
        color = IndexedColor(2, palette.colors8_dark)
        self.assertIs(type(color.rgb_tuple), tuple)
        self.assertEqual(color.rgb_tuple, palette.colors8_dark.rgb_tuple(2))

    def test_name(self):
        self.assertEqual(
            IndexedColor(2, palette.colors8_dark).name,
            palette.colors8_dark.name(2)
        )

    def test_cmp(self):
        a = IndexedColor(2, palette.colors8_bright)
        b = IndexedColor(2, palette.colors8_bright)
        c = IndexedColor(3, palette.colors8_bright)
        d = IndexedColor(2, palette.colors8_dark)
        e = IndexedColor(10, palette.colors16)

        self.assertTrue(a == b)
        self.assertFalse(a == c)
        self.assertFalse(a == d)
        self.assertFalse(a == None)
        self.assertTrue(a == e)

        self.assertFalse(a != b)
        self.assertTrue(a != c)
        self.assertTrue(a != d)
        self.assertFalse(a != e)
        self.assertTrue(a != None)

        self.assertEqual(hash(a), hash(b))
        self.assertNotEqual(hash(a), hash(c))
        self.assertNotEqual(hash(a), hash(d))
        self.assertNotEqual(hash(a), hash(None))
        self.assertNotEqual(hash(a), hash(e)) # Should this result equal?


class TestRgbColor(test_common.TestCase):
    def test_hex_rgb(self):
        self.assertEqual(RgbColor(0xf1, 0x2, 0x34).hex(), "#f10234")

    def test_ctor(self):
        c = RgbColor(1, 2, 3)
        self.assertEqual(c.r, 1)
        self.assertEqual(c.g, 2)
        self.assertEqual(c.b, 3)

    def test_rgb(self):
        c = RgbColor(1, 2, 3)
        self.assertIs(type(c.rgb), RgbColor)
        self.assertEqual(c.rgb, c)

    def test_rgb_tuple(self):
        c = RgbColor(1, 2, 3)
        self.assertIs(type(c.rgb_tuple), tuple)
        self.assertEqual(c.rgb_tuple, (1, 2, 3))

    def test_name(self):
        c = RgbColor(1, 2, 3)
        self.assertEqual(c.name, c.hex())
        c.name = "foocolor"
        self.assertEqual(c.name, "foocolor")
        del c.name
        self.assertEqual(c.name, c.hex())

    def test_cmp(self):
        a = RgbColor(1, 2, 3)
        b = RgbColor(1, 2, 3, "foocolor")
        c = RgbColor(9, 2, 3)
        d = RgbColor(1, 9, 3)
        e = RgbColor(1, 2, 9)
        p = palette.Palette(
            ("a", (1,2,3)),
            ("b", (3,2,1)),
        )
        f = IndexedColor(0, p)
        g = IndexedColor(1, p)

        self.assertTrue(a == b)
        self.assertFalse(a == c)
        self.assertFalse(a == d)
        self.assertFalse(a == e)
        self.assertTrue(a == f)
        self.assertFalse(a == g)
        self.assertFalse(a == None)

        self.assertFalse(a != b)
        self.assertTrue(a != c)
        self.assertTrue(a != d)
        self.assertTrue(a != e)
        self.assertFalse(a != f)
        self.assertTrue(a != g)
        self.assertTrue(a != None)

        self.assertEqual(hash(a), hash(b))
        self.assertNotEqual(hash(a), hash(c))
        self.assertNotEqual(hash(a), hash(d))
        self.assertNotEqual(hash(a), hash(e))
        self.assertNotEqual(hash(a), hash(f)) # Should be equal?
        self.assertNotEqual(hash(a), hash(g))
        self.assertNotEqual(hash(a), hash(None))

    def test_copy(self):
        color = RgbColor(1, 2, 3, "foocolor")
        copy = color.copy()
        self.assertEqual(color, copy)
        self.assertEqual(color.name, copy.name)
        self.assertIsNot(color, copy)

    def test_tuple_float(self):
        self.assertTupleEqual(RgbColor(0, 0, 0).rgb_float(), (0.0, 0.0, 0.0))
        self.assertTupleEqual(RgbColor(255, 0, 0).rgb_float(), (1.0, 0.0, 0.0))
        self.assertTupleEqual(RgbColor(255, 255, 0).rgb_float(), (1.0, 1.0, 0.0))
        self.assertTupleEqual(RgbColor(255, 255, 255).rgb_float(), (1.0, 1.0, 1.0))

    def test_lab(self):
        self.assertTupleEqual(RgbColor(0, 0, 0).lab(), (0, 0, 0))
        self.assertTupleEqual(RgbColor(255, 0, 0).lab(), (53, 80, 67))
        self.assertTupleEqual(RgbColor(255, 255, 255).lab(), (100, 0, 0))

    def test_distance(self):
        a = RgbColor(71, 92, 43)
        self.assertAlmostEqual(a.distance(a), 0)

        b = RgbColor(73, 92, 43)
        c = RgbColor(74, 92, 43)
        self.assertLess(a.distance(b), a.distance(c))

        d = RgbColor(67, 103, 28)
        e = RgbColor(129, 92, 43)
        self.assertLess(a.distance(d), a.distance(e))

    def test_to_indexed(self):
        red = RgbColor(255, 0, 0)
        red_ish = RgbColor(200, 78, 24)
        cyan_ish = RgbColor(93, 234, 194)

        self.assertRaises(Exception, lambda: red.to_indexed(palette.Palette()))

        self.assertEqual(red.to_indexed(palette.colors8_dark).index, 1)
        self.assertEqual(red_ish.to_indexed(palette.colors8_dark).index, 1)
        self.assertEqual(cyan_ish.to_indexed(palette.colors8_dark).index, 6)

    def test_from_hsv(self):
        self.assertEquals(RgbColor.from_hsv(0, 0, 0).rgb_tuple, (0, 0, 0))
        self.assertEquals(RgbColor.from_hsv(0.5, 0, 0).rgb_tuple, (0, 0, 0))
        self.assertEquals(RgbColor.from_hsv(0.5, 1, 0).rgb_tuple, (0, 0, 0))

        self.assertEquals(RgbColor.from_hsv(0, 1, 1).rgb_tuple, (255, 0, 0))
        self.assertEquals(RgbColor.from_hsv(-1, 1, 1).rgb_tuple, (255, 0, 0))
        self.assertEquals(RgbColor.from_hsv(1, 1, 1).rgb_tuple, (255, 0, 0))
        self.assertEquals(RgbColor.from_hsv(0.5, 1, 1).rgb_tuple, (0, 255, 255))
        self.assertEquals(RgbColor.from_hsv(1.5, 1, 1).rgb_tuple, (0, 255, 255))

        self.assertEquals(RgbColor.from_hsv(0/6.0, 1, 1).rgb_tuple, (255, 0, 0))
        self.assertEquals(RgbColor.from_hsv(1/6.0, 1, 1).rgb_tuple, (255, 255, 0))
        self.assertEquals(RgbColor.from_hsv(2/6.0, 1, 1).rgb_tuple, (0, 255, 0))
        self.assertEquals(RgbColor.from_hsv(3/6.0, 1, 1).rgb_tuple, (0, 255, 255))
        self.assertEquals(RgbColor.from_hsv(4/6.0, 1, 1).rgb_tuple, (0, 0, 255))
        self.assertEquals(RgbColor.from_hsv(5/6.0, 1, 1).rgb_tuple, (255, 0, 255))
        self.assertEquals(RgbColor.from_hsv(6/6.0, 1, 1).rgb_tuple, (255, 0, 0))

    def test_hsv(self):
        self.assertEquals(RgbColor(255, 0, 0).hsv(), (0/6.0, 1, 1))
        self.assertEquals(RgbColor(255, 255, 0).hsv(), (1/6.0, 1, 1))
        self.assertEquals(RgbColor(0, 255, 0).hsv(), (2/6.0, 1, 1))
        self.assertEquals(RgbColor(0, 255, 255).hsv(), (3/6.0, 1, 1))
        self.assertEquals(RgbColor(0, 0, 255).hsv(), (4/6.0, 1, 1))
        self.assertEquals(RgbColor(255, 0, 255).hsv(), (5/6.0, 1, 1))
        self.assertEquals(RgbColor(255, 0, 0).hsv(), (0/6.0, 1, 1))

        self.assertEquals(RgbColor(255, 255, 255).hsv(), (0, 0, 1))
        self.assertEquals(RgbColor(0, 0, 0).hsv(), (0, 0, 0))


test_common.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
import six
import xml.etree.cElementTree as ElementTree

from . import factory
from .. import tree


class XmlFormatter(object):
    flat = False

    def document_element(self, doc):
        xml_doc = ElementTree.Element("document")
        xml_doc.attrib["name"] = doc.name
        if doc.metadata:
            metadata = ElementTree.SubElement(xml_doc, "metadata")
            for key, val in six.iteritems(doc.metadata):
                element = ElementTree.SubElement(metadata, key)
                element.text = val
        for layer in doc.layers:
            xml_doc.append(self.layer_element(layer))
        return xml_doc

    def layer_element(self, layer):
        if isinstance(layer, tree.Layer):
            xml_layer = ElementTree.Element("layer")
            xml_layer.attrib["color"] = self.color(layer.color)
            xml_layer.text = layer.text
            return xml_layer
        elif isinstance(layer, tree.FreeColorLayer):
            xml_layer = ElementTree.Element("layer")
            for pos in sorted(layer.matrix.keys(), key=lambda a: (a[1], a[0])):
                char_xml = ElementTree.SubElement(xml_layer, "char")
                char_xml.attrib["color"] = self.color(layer.matrix[pos][1])
                char_xml.attrib["x"] = str(pos[0])
                char_xml.attrib["y"] = str(pos[1])
                char_xml.text = layer.matrix[pos][0]
            return xml_layer
        else:
            raise TypeError("Expected layer type")

    def write_element(self, element, output):
        ElementTree.ElementTree(element).write(output, "utf-8", True)

    def document(self, doc, output):
        self.write_element(self.document_element(doc), output)

    def layer(self, layer, output):
        self.write_element(self.layer_element(layer), output)

    def color(self, color):
        if color is None:
            return ""
        return color.name


factory.register(XmlFormatter(), "xml")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import re
import six

from . import factory
from .ansi import AnsiFormatter


_default_replacements = {
    "\n": "\\n",
    "\"": "\\\"",
    "\\": "\\\\",
    "\x1b": "\\x1b",
}


class AnsiSourceFormatter(object):
    flat = True

    def __init__(self, prefix='"', suffix='"\n',
                 replacements=_default_replacements, *args, **kwargs):
        self.prefix = prefix
        self.suffix = suffix
        self._ansi_formatter = AnsiFormatter(*args, **kwargs)
        self._repl_dict = replacements
        self._repl_search = re.compile("|".join(map(re.escape, replacements.keys())))
        self._repl_replace = lambda match: self._repl_dict[match.group()]

    def _wrap_ansi(self, func, object, output):
        output.write(self.prefix)
        buffer = six.StringIO()
        func(object, buffer)
        output.write(self._replace(buffer.getvalue()))
        output.write(self.suffix)

    def _replace(self, string):
        return self._repl_search.sub(self._repl_replace, string)

    def document(self, doc, output):
        self._wrap_ansi(self._ansi_formatter.document, doc, output)

    def layer(self, layer, output):
        self._wrap_ansi(self._ansi_formatter.layer, layer, output)

    def color(self, color):
        return (
            self.prefix +
            self._replace(self._ansi_formatter.color(color))  +
            self.suffix
        )

AnsiSourceFormatter.builtins = {
    "bash": AnsiSourceFormatter(
        "#!/bin/bash\nread -r -d '' Heredoc_var <<'Heredoc_var'\n\\x1b[0m",
        "\\x1b[0m\nHeredoc_var\necho -e \"$Heredoc_var\"\n",
        {"\\": "\\\\", "\x1b": "\\x1b"}
    ),
    "python": AnsiSourceFormatter(
        "print('''",
        "''')\n",
        {"\\": "\\\\", "\'": "\\\'", "\x1b": "\\x1b"}
    ),
    "perl": AnsiSourceFormatter(
        "print \"",
        "\";\n",
        {"\\": "\\\\", "\"": "\\\"", "\x1b": "\\x1b",
         "$": "\\$", "@": "\\@", "%": "\\%"}
    ),
    "php": AnsiSourceFormatter(
        "<?php\nprint \"",
        "\";\n",
        {"\\": "\\\\", "\"": "\\\"", "\x1b": "\\x1b", "$": "\\$"}
    ),
}


factory.register(AnsiSourceFormatter.builtins["bash"], "sh")
factory.register(AnsiSourceFormatter.builtins["python"], "py")
factory.register(AnsiSourceFormatter.builtins["perl"], "pl")
factory.register(AnsiSourceFormatter.builtins["php"], "php")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


def next_object(array, current, delta):
    if not array:
        return current
    if current not in array:
        index = delta - 1 if delta > 0 else delta
    else:
        index = array.index(current) + delta
    return array[index % len(array)]



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import test_common

from patsi.ansi import *


class TestAnsiCode(test_common.TestCase):
    def test_ctor(self):
        a = AnsiCode("m")
        self.assertEqual(a.terminator, "m")
        self.assertEqual(a.args(), [])
        self.assertEqual(a.private, "")

        a = AnsiCode("m", [1])
        self.assertEqual(a.terminator, "m")
        self.assertEqual(a.args(), [1])
        self.assertEqual(a.private, "")

        a = AnsiCode("m", [1], "?")
        self.assertEqual(a.terminator, "m")
        self.assertEqual(a.args(), [1])
        self.assertEqual(a.private, "?")

    def test_repr(self):
        a = AnsiCode("m", [1, 2, 3], "?")
        self.assertEqual(repr(a), "\x1b[?1;2;3m")

    def test_parse(self):
        tests = [
            "\x1b[K",
            "\x1b[1J",
            "\x1b[?25h",
            "\x1b[31;1m",
        ]
        for test in tests:
            a = AnsiCode.parse(test)
            self.assertTrue(isinstance(a, AnsiCode))
            self.assertEqual(repr(a), test)
        self.assertRaises(Exception, lambda: AnsiCode.parse("Foo"))

    def test_split(self):
        split = list(AnsiCode.split("Hello\x1b[31mWorld\x1b[H\x1b[0mFoo"))
        self.assertEqual(split[0], "Hello")
        self.assertTrue(isinstance(split[1], AnsiCode))
        self.assertEqual(split[2], "World")
        self.assertTrue(isinstance(split[3], AnsiCode))
        self.assertTrue(isinstance(split[4], AnsiCode))
        self.assertEqual(split[5], "Foo")
        self.assertEqual(list(AnsiCode.split("")), [])

    def test_compare(self):
        self.assertTrue(AnsiCode("m", ["1", "2", "3"]) ==
                        AnsiCode.parse("\x1b[1;2;3m"))

        self.assertTrue(AnsiCode("q", ["1", "2", "3"]) !=
                        AnsiCode.parse("\x1b[1;2;3m"))

        self.assertTrue(AnsiCode("m", ["1", "5", "3"]) !=
                        AnsiCode.parse("\x1b[1;2;3m"))

        self.assertTrue(AnsiCode("m", ["1", "2", "3"], "?") !=
                        AnsiCode.parse("\x1b[1;2;3m"))


class TestCursorPosition(test_common.TestCase):
    def test_parse(self):
        a = AnsiCode.parse("\x1b[3;4H")
        self.assertTrue(isinstance(a, CursorPosition))
        self.assertEqual(a.row, 3)
        self.assertEqual(a.column, 4)

        a = AnsiCode.parse("\x1b[3;4f")
        self.assertTrue(isinstance(a, CursorPosition))
        self.assertEqual(a.row, 3)
        self.assertEqual(a.column, 4)

        a = AnsiCode.parse("\x1b[H")
        self.assertTrue(isinstance(a, CursorPosition))
        self.assertEqual(a.row, 1)
        self.assertEqual(a.column, 1)

    def test_ctor(self):
        a = CursorPosition(3, 4)
        self.assertEqual(a.row, 4)
        self.assertEqual(a.column, 3)

        a = CursorPosition(x=3, y=4)
        self.assertEqual(a.row, 4)
        self.assertEqual(a.column, 3)

        a = CursorPosition(column=3, row=4)
        self.assertEqual(a.row, 4)
        self.assertEqual(a.column, 3)

        a = CursorPosition()
        self.assertEqual(a.row, 1)
        self.assertEqual(a.column, 1)

    def test_args(self):
        self.assertEqual(CursorPosition(row=3, column=4).args(), [3, 4])


class TestGraphicRendition(test_common.TestCase):
    def test_alias(self):
        self.assertEqual(SGR, GraphicRendition)

    def test_parse(self):
        a = AnsiCode.parse("\x1b[m")
        self.assertTrue(isinstance(a, GraphicRendition))
        self.assertEqual(a.flags, [])

        a = AnsiCode.parse("\x1b[1;2;3;4m")
        self.assertTrue(isinstance(a, GraphicRendition))
        self.assertEqual(a.flags, [1, 2, 3, 4])

    def test_ctor_simple(self):
        self.assertEqual(SGR().flags, [])
        self.assertEqual(SGR(1, 2, 3).flags, [1, 2, 3])
        self.assertEqual(SGR([1, 2, 3]).flags, [1, 2, 3])
        self.assertEqual(SGR(1).flags, [1])

    def test_ctor_color(self):
        a = SGR([31, 42, 93, 104])
        self.assertTrue(isinstance(a.flags[0], SGR.Color))
        self.assertEqual(a.flags[0].color, SGR.Color.Red)
        self.assertEqual(a.flags[0].background, False)
        self.assertEqual(a.flags[0].bright, False)
        self.assertEqual(repr(a.flags[0]), "31")

        self.assertTrue(isinstance(a.flags[1], SGR.Color))
        self.assertEqual(a.flags[1].color, SGR.Color.Green)
        self.assertEqual(a.flags[1].background, True)
        self.assertEqual(a.flags[1].bright, False)
        self.assertEqual(repr(a.flags[1]), "42")

        self.assertTrue(isinstance(a.flags[2], SGR.Color))
        self.assertEqual(a.flags[2].color, SGR.Color.Yellow)
        self.assertEqual(a.flags[2].background, False)
        self.assertEqual(a.flags[2].bright, True)
        self.assertEqual(repr(a.flags[2]), "93")

        self.assertTrue(isinstance(a.flags[3], SGR.Color))
        self.assertEqual(a.flags[3].color, SGR.Color.Blue)
        self.assertEqual(a.flags[3].background, True)
        self.assertEqual(a.flags[3].bright, True)
        self.assertEqual(repr(a.flags[3]), "104")

    def test_ctor_color256(self):
        a = SGR([38, 5, 69, 48, 5, 42])
        self.assertEqual(len(a.flags), 2)

        self.assertTrue(isinstance(a.flags[0], SGR.Color256))
        self.assertEqual(a.flags[0].color, 69)
        self.assertEqual(a.flags[0].background, False)
        self.assertEqual(repr(a.flags[0]), "38;5;69")

        self.assertTrue(isinstance(a.flags[1], SGR.Color256))
        self.assertEqual(a.flags[1].color, 42)
        self.assertEqual(a.flags[1].background, True)
        self.assertEqual(repr(a.flags[1]), "48;5;42")

    def test_ctor_color_rgb(self):
        a = SGR([38, 2, 1, 2, 3, 48, 2, 1, 2, 3])
        self.assertEqual(len(a.flags), 2)

        self.assertTrue(isinstance(a.flags[0], SGR.ColorRGB))
        self.assertEqual(a.flags[0].r, 1)
        self.assertEqual(a.flags[0].g, 2)
        self.assertEqual(a.flags[0].b, 3)
        self.assertEqual(a.flags[0].background, False)
        self.assertEqual(repr(a.flags[0]), "38;2;1;2;3")

        self.assertTrue(isinstance(a.flags[1], SGR.ColorRGB))
        self.assertEqual(a.flags[1].r, 1)
        self.assertEqual(a.flags[1].g, 2)
        self.assertEqual(a.flags[1].b, 3)
        self.assertEqual(a.flags[1].background, True)
        self.assertEqual(repr(a.flags[1]), "48;2;1;2;3")

    def test_ctor_color_invalid(self):
        # TODO Is this the right behaviour or should it eat the mismatched flags?
        self.assertEqual(len(SGR([98, 5, 2, 3]).flags), 3)
        self.assertEqual(len(SGR([38, 2, 1]).flags), 2)
        self.assertEqual(len(SGR([48, 2, 1]).flags), 2)
        self.assertEqual(len(SGR([108, 2, 1]).flags), 2)

    def test_background(self):
        color = SGR.Red
        bg = SGR.Background(color)
        self.assertEqual(bg.color, color.color)
        self.assertTrue(bg.background)
        self.assertFalse(color.background)

    def test_reverse(self):
        reversible = [
            (SGR.Bold, 22),
            (SGR.Italic, 23),
            (SGR.Underline, 24),
        ]
        for a, b in reversible:
            self.assertEqual(SGR.reverse(a), b)
            self.assertEqual(SGR.reverse(b), a)
        self.assertEqual(SGR.reverse(41), 49)
        self.assertEqual(SGR.reverse(31), 39)
        self.assertEqual(SGR.reverse(SGR.Red), 39)
        self.assertEqual(SGR.reverse(91), 39)
        self.assertEqual(SGR.reverse(SGR.Background(SGR.ColorRGB(1,2,3))), 49)
        self.assertEqual(SGR.reverse(1234), 0)

    def test_args(self):
        self.assertEqual(SGR([1, 2, 3]).args(), [1, 2, 3])

    def test_flag_values(self):
        pass


class TestCharMover(test_common.TestCase):
    def test_move(self):
        m = CharMover(0, 0)
        self.assertFalse(m.moved)
        m.move(2, 3)
        self.assertTrue(m.moved)
        self.assertEqual(m.x, 2)
        self.assertEqual(m.y, 3)

    def test_loop(self):
        m = CharMover(0, 0)
        string = "   ab\nc\td\re\vf"
        gen = m.loop(string)

        self.assertFalse(m.moved)
        self.assertEqual(next(gen), 'a')
        self.assertEqual((m.x, m.y), (3, 0))
        self.assertTrue(m.moved)

        self.assertEqual(next(gen), 'b')
        self.assertEqual((m.x, m.y), (4, 0))
        self.assertFalse(m.moved)

        self.assertEqual(next(gen), 'c')
        self.assertEqual((m.x, m.y), (0, 1))
        self.assertTrue(m.moved)

        self.assertEqual(next(gen), 'd')
        self.assertEqual((m.x, m.y), (5, 1))
        self.assertTrue(m.moved)

        self.assertEqual(next(gen), 'e')
        self.assertEqual((m.x, m.y), (0, 1))
        self.assertTrue(m.moved)

        self.assertEqual(next(gen), 'f')
        self.assertEqual((m.x, m.y), (2, 2))
        self.assertTrue(m.moved)


class TestAnsiRenderer(test_common.StringOutputTestCase):
    def setUp(self):
        super(TestAnsiRenderer, self).setUp()
        self.renderer = AnsiRenderer(self.output)

    def test_ctor(self):
        self.assertIs(self.renderer.output, self.output)

    def test_context_nocursor(self):
        with self.renderer.context_nocursor():
            self._check_data(common.hide_cursor)
            self._clear_data()
        self._check_data(common.show_cursor)

    def test_context_alt_screen(self):
        with self.renderer.context_alt_screen():
            self._check_data(common.alt_screen)
            self._clear_data()
        self._check_data(common.main_screen)

    def test_context_sgr(self):
        self.renderer.ansi(SGR(SGR.Bold))
        with self.renderer.context_sgr(SGR.Underline):
            self._check_data(SGR(SGR.Bold), SGR(SGR.Underline))
        self._check_data(SGR(SGR.Bold), SGR(SGR.Underline), SGR(SGR.reverse(SGR.Underline)))

        self._clear_data()
        self.renderer.ansi(SGR(SGR.Bold))
        with self.renderer.context_sgr(SGR.Underline, hard_reset=True):
            self._check_data(SGR(SGR.Bold), SGR(SGR.Underline))
        self._check_data(SGR(SGR.Bold), SGR(SGR.Underline), SGR(0))

    def test_rendering_context(self):
        with self.renderer.rendering_context():
            self._check_data(common.alt_screen, common.hide_cursor,
                             common.clear_screen, SGR(SGR.Reset))
            self._clear_data()
            raise KeyboardInterrupt
        self._check_data(common.show_cursor, common.main_screen)

    def test_keyboard_interrupt(self):
        with keyboard_interrupt():
            raise KeyboardInterrupt
        self.assertTrue(True)

    def test_ansi(self):
        self.renderer.ansi(SGR(SGR.Bold))
        self._check_data(SGR(SGR.Bold))

        self._clear_data()
        self.renderer.ansi("m", [1])
        self._check_data(SGR(SGR.Bold))

    def test_move_cursor(self):
        self.renderer.move_cursor(2, 3)
        self._check_data(CursorPosition(2, 3))

        self._clear_data()
        self.renderer.move_cursor()
        self._check_data(CursorPosition())

    def test_clear(self):
        self.renderer.clear()
        self._check_data(common.clear_screen)

    def test_write(self):
        self.renderer.write("Hello")
        self.assertEqual(self.output.getvalue(), "Hello")

    def test_overlay(self):
        self.renderer.overlay("abc\nd\x1b[31me")
        self.assertEqual(self.output.getvalue(), "abc\x1b[2;1Hd\x1b[31me")


test_common.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import six

from . import factory
from .. import tree


class TextFormatter(object):
    flat = True

    def document(self, doc, output):
        self.layer(doc.flattened(), output)

    def layer(self, layer, output):
        if isinstance(layer, tree.Layer):
            output.write(layer.text)
        elif isinstance(layer, tree.FreeColorLayer):
            for y in six.moves.range(layer.height):
                for x in six.moves.range(layer.width):
                    char, color = layer.matrix.get((x, y), (" ", None))
                    output.write(char)
                output.write("\n")
        else:
            raise TypeError("Expected layer type")

    def color(self, color):
        return ""


factory.register(TextFormatter(), "txt")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import six

from . import factory
from .. import tree
from .. import color
from .. import palette


class IrcFormatter(object):
    flat = True
    colors = [
        "01", "05", "03", "07", "02", "06", "10", "15",
        "14", "04", "09", "08", "12", "13", "11", "00"
    ]

    def document(self, doc, output):
        self.layer(doc.flattened(), output)

    def layer(self, layer, output):
        if isinstance(layer, tree.Layer):
            for line in layer.text.splitlines(True):
                output.write(self.color(layer.color) + line)
        elif isinstance(layer, tree.FreeColorLayer):
            for y in six.moves.range(layer.height):
                prev_color = None
                for x in six.moves.range(layer.width):
                    char, col = layer.matrix.get((x, y), (" ", color.UnchangedColor))
                    if col is not color.UnchangedColor and col != prev_color:
                        prev_color = col
                        output.write(self.color(col))
                    output.write(char)
                output.write("\n")
        else:
            raise TypeError("Expected layer type")

    def color(self, col):
        if col is None:
            return "\x0f"
        elif isinstance(col, color.IndexedColor):
            if len(col.palette) == 8:
                bright = 8 if getattr(col.palette, "bright", False) else 0
                code = self.colors[col.index + bright]
            elif len(col.palette) == 16:
                code = self.colors[col.index]
            else:
                code = self.colors[col.rgb.to_indexed(palette.colors16).index]
        elif isinstance(col, color.RgbColor):
            code = self.colors[col.to_indexed(palette.colors16).index]
        else:
            raise TypeError("Expected document color")

        return "\x03%s,01" % code


factory.register(IrcFormatter(), "irc")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import six

from . import factory
from .. import tree
from .. import color
from .. import palette
from ...ansi import SGR


class AnsiFormatter(object):
    class Features(object):
        STANDARD    = 0     # Standard, only 8 colors supported
        BRIGHT_BOLD = 1     # Colors can be made bright by setting the bold flag
        BRIGHT      = 2     # Supports 9x codes for bright colors
        EXT_PALETTE = 256   # Supports 256-color palette (Code 38;5;*)
        RGB         = 4     # Supports 24 bit RGB colors (Code 38;2;r;g;b)
        ALL         = 0xFFF # All features are supported

        @classmethod
        def constants_dict(cls):
            if not hasattr(cls, "_constants_dict"):
                cls._constants_dict = {
                    name: value
                    for name, value in vars(cls).iteritems()
                    if name.isupper()
                }
            return cls._constants_dict

        @classmethod
        def feature_from_name(cls, name):
            return cls.constants_dict().get(name.strip().upper(), 0)

        def __init__(self, features=ALL):
            if type(features) is int:
                self.features = features
            elif isinstance(features, AnsiFormatter.Features):
                self.features = features.features
            else:
                if type(features) is str or type(features) is unicode:
                    features = features.split(", ")

                self.features = 0
                for feature in features:
                    if type(feature) is int:
                        self.features |= feature
                    else:
                        self.features |= self.feature_from_name(feature)

        def has_feature(self, feature):
            if type(feature) is str or type(feature) is unicode:
                feature = self.feature_from_name(feature)
            return (self.features & feature) == feature

    flat = True
    features = Features()

    def document(self, doc, output):
        self.layer(doc.flattened(), output)

    def layer(self, layer, output):
        if isinstance(layer, tree.Layer):
            for line in layer.text.splitlines(True):
                output.write(self.color(layer.color) + line)
        elif isinstance(layer, tree.FreeColorLayer):
            for y in six.moves.range(layer.height):
                prev_color = None
                for x in six.moves.range(layer.width):
                    char, col = layer.matrix.get((x, y), (" ", color.UnchangedColor))
                    if col is not color.UnchangedColor and col != prev_color:
                        prev_color = col
                        output.write(self.color(col))
                    output.write(char)
                output.write("\n")
        else:
            raise TypeError("Expected layer type")

    def color(self, col):
        return str(self._color_sgr(col))

    def _color_sgr(self, col):
        if col is None:
            ansi = SGR(SGR.ResetColor)
            self._bold = False
        elif isinstance(col, color.RgbColor):
            ansi = self._rgb_to_ansi(col)
        elif isinstance(col, color.IndexedColor):
            if len(col.palette) == 8:
                if getattr(col.palette, "bright", False):
                    ansi = self._16_to_ansi(col.index | 8)
                else:
                    ansi = self._8_to_ansi(col.index)
            elif len(col.palette) == 16:
                ansi = self._16_to_ansi(col.index)
            else:
                ansi = self._256_to_ansi(col)
        else:
            raise TypeError("Expected document color")

        if SGR.Bold in ansi.flags:
            self._bold = True
        elif getattr(self, "_bold", False):
            self._bold = False
            ansi.flags.append(SGR.reverse(SGR.Bold))
        return ansi

    def _rgb_to_ansi(self, col):
        if self.features.has_feature(AnsiFormatter.Features.RGB):
            return SGR(SGR.ColorRGB(col.r, col.g, col.b))
        if self.features.has_feature(AnsiFormatter.Features.EXT_PALETTE):
            return self._256_to_ansi(col.to_indexed(palette.colors256))
        return self._16_to_ansi(col.to_indexed(palette.colors16).index)

    def _8_to_ansi(self, index):
        return SGR(SGR.Color(index))

    def _16_to_ansi(self, index):
        bright = index > 7
        index = index & 7
        if not bright:
            return self._8_to_ansi(index)
        if self.features.has_feature(AnsiFormatter.Features.BRIGHT):
            return SGR(SGR.Color(index, False, True))
        if self.features.has_feature(AnsiFormatter.Features.BRIGHT_BOLD):
            return SGR(SGR.Color(index), SGR.Bold)
        return SGR(SGR.Color(index))

    def _256_to_ansi(self, col):
        if self.features.has_feature(AnsiFormatter.Features.EXT_PALETTE):
            return SGR(SGR.Color256(col.index))
        return self._rgb_to_ansi(col.rgb)

factory.register(AnsiFormatter(), "ansi")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import inspect
import unittest
from unittest import TestCase
import six


sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


class StringOutputTestCase(unittest.TestCase):
    def setUp(self):
        self.output = six.StringIO()

    def _get_data(self):
        return self.output.getvalue()

    def _clear_data(self):
        self.output.truncate(0)
        self.output.seek(0)

    def _check_data(self, *args):
        self.assertEqual(
            self._get_data(),
            "".join(str(arg) for arg in args)
        )


class BytesOutputTestCase(StringOutputTestCase):
    def setUp(self):
        self.output = six.BytesIO()

    def _check_data(self, *args):
        self.assertEqual(
            self._get_data(),
            b"".join(
                arg if type(arg) is bytes else bytes(str(arg), "utf-8")
                for arg in args
            )
        )


class MockFile(object):
    def __init__(self, wrapped=six.StringIO()):
        self._wrapped = wrapped

    def __getattr__(self, name):
        if name != "_wrapped" and name not in vars(self):
            return getattr(self._wrapped, name)
        return super(MockFile, self).__getattr__(name)

    def __delattr__(self, name):
        if name != "_wrapped" and name not in vars(self):
            return delattr(self._wrapped, name)

    def __setattr__(self, name, value):
        if name != "_wrapped" and name not in vars(self):
            return setattr(self._wrapped, name, value)
        return super(MockFile, self).__setattr__(name, value)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def open(self, file, mode="r"):
        self.name = file
        self.mode = mode
        return self

    def reset(self, contents=""):
        self._wrapped = six.StringIO(contents)


def main():
    if inspect.getmodule(inspect.stack()[1][0]).__name__ == "__main__":
        unittest.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
import json

from . import factory
from .. import tree
from ._utils import string_to_color


class JsonLoader(object):
    def _load_json_layer(self, json_dict):
        if "chars" in json_dict:
            layer = tree.FreeColorLayer()
            for char in json_dict["chars"]:
                layer.set_char(
                    int(char["x"]),
                    int(char["y"]),
                    char["char"],
                    string_to_color(char.get("color", ""))
                )
            return layer

        return tree.Layer(
            json_dict["text"],
            string_to_color(json_dict.get("color", ""))
        )

    def load_json(self, json_dict):
        return tree.Document(
            json_dict.get("name", ""),
            [self._load_json_layer(layer) for layer in json_dict.get("layers", [])],
            json_dict.get("metadata", {}),
        )

    def load_string(self, string):
        return self.load_json(json.loads(string))

    def load_file(self, file):
        if type(file) is str:
            with open(file) as f:
                return self.load_file(f)
        return self.load_json(json.load(file))


factory.register(JsonLoader(), "json")



#!/usr/bin/env python
#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


class Point(object):
    """
    Point in a 2D plane
    """

    def __init__(self, *args, **kwargs):
        if len(args) == 2:
            if kwargs.get("reversed", False):
                args = reversed(args)
            self.x, self.y = args
        elif len(args) == 1:
            self.x, self.y = args[0]
        else:
            self.x = kwargs.get("x", 0)
            self.y = kwargs.get("y", 0)

    def __iadd__(self, other):
        self.x += other.x
        self.y += other.y
        return self

    def __add__(self, other):
        return Point(
            self.x + other.x,
            self.y + other.y
        )

    def __isub__(self, other):
        self.x -= other.x
        self.y -= other.y
        return self

    def __sub__(self, other):
        return Point(
            self.x - other.x,
            self.y - other.y
        )

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return not (self == other)

    def interpolate(self, other, factor=0.5):
        def lerp(a, b, factor):
            return int(round(a * (1 - factor) + b * factor))
        return Point(
            lerp(self.x, other.x, factor),
            lerp(self.y, other.y, factor)
        )

    def __repr__(self):
        return "(%s, %s)" % (self.x, self.y)

    def __len__(self):
        return 2

    def __getitem__(self, key):
        if key == 0 or key == "x":
            return self.x
        if key == 1 or key == "y":
            return self.y
        raise KeyError(key)

    def __iter__(self):
        yield self.x
        yield self.y


class Rect(object):
    def __init__(self, **kwargs):
        if "pos" in kwargs and "size" in kwargs:
            self._top_left = kwargs["pos"]
            self._bottom_right = self._top_left + kwargs["size"]
        elif "top_left" in kwargs and "bottom_right" in kwargs:
            self._top_left = kwargs["top_left"]
            self._bottom_right = kwargs["bottom_right"]
        elif ( "x" in kwargs and "y" in kwargs and
                "width" in kwargs and "height" in kwargs ):
            self._top_left = Point(kwargs["x"], kwargs["y"])
            self._bottom_right = self._top_left + \
                Point(kwargs["width"], kwargs["height"])
        elif ( "x1" in kwargs and "y1" in kwargs and
                "x2" in kwargs and "y2" in kwargs ):
            self._top_left = Point(kwargs["x1"], kwargs["y1"])
            self._bottom_right = Point(kwargs["x2"], kwargs["y2"])
        elif ( "top" in kwargs and "left" in kwargs and
                "bottom" in kwargs and "right" in kwargs ):
            self._top_left = Point(kwargs["top"], kwargs["left"])
            self._bottom_right = Point(kwargs["bottom"], kwargs["right"])
        else:
            self._top_left = Point()
            self._bottom_right = Point()

    def contains(self, point, inclusive=True):
        if inclusive:
            return self._top_left.x <= point.x <= self._bottom_right.x and \
                   self._top_left.y <= point.y <= self._bottom_right.y
        return self._top_left.x < point.x < self._bottom_right.x and \
               self._top_left.y < point.y < self._bottom_right.y

    @property
    def top(self):
        return self._top_left.y

    @property
    def left(self):
        return self._top_left.x

    @property
    def bottom(self):
        return self._bottom_right.y

    @property
    def right(self):
        return self._bottom_right.x

    @property
    def top_left(self):
        return Point(self.left, self.top)

    @property
    def top_right(self):
        return Point(self.right, self.top)

    @property
    def bottom_left(self):
        return Point(self.left, self.bottom)

    @property
    def bottom_right(self):
        return Point(self.right, self.bottom)

    @property
    def center(self):
        return self._top_left.interpolate(self._bottom_right)

    @property
    def x1(self):
        return self._top_left.x

    @property
    def y1(self):
        return self._top_left.y

    @property
    def x2(self):
        return self._bottom_right.x

    @property
    def y2(self):
        return self._bottom_right.y

    @property
    def x(self):
        return self._top_left.x

    @property
    def y(self):
        return self._top_left.y

    @property
    def width(self):
        return self._bottom_right.x - self._top_left.x

    @property
    def height(self):
        return self._bottom_right.y - self._top_left.y

    @property
    def pos(self):
        return Point(
            self._top_left.x,
            self._top_left.y
        )

    @property
    def size(self):
        return self._bottom_right - self._top_left

    @top_left.setter
    def top_left(self, point):
        self._top_left.x = point.x
        self._top_left.y = point.y

    @bottom_right.setter
    def bottom_right(self, point):
        self._bottom_right.x = point.x
        self._bottom_right.y = point.y

    @top_right.setter
    def top_right(self, point):
        self._bottom_right.x = point.x
        self._top_left.y = point.y

    @bottom_left.setter
    def bottom_left(self, point):
        self._top_left.x = point.x
        self._bottom_right.y = point.y

    @left.setter
    def left(self, val):
        self._top_left.x = val

    @top.setter
    def top(self, val):
        self._top_left.y = val

    @bottom.setter
    def bottom(self, val):
        self._bottom_right.y = val

    @right.setter
    def right(self, val):
        self._bottom_right.x = val

    @x.setter
    def x(self, val):
        width = self.width
        self._top_left.x = val
        self._bottom_right.x = val + width

    @y.setter
    def y(self, val):
        height = self.height
        self._top_left.y = val
        self._bottom_right.y = val + height

    @width.setter
    def width(self, val):
        self._bottom_right.x = self._top_left.x + val

    @height.setter
    def height(self, val):
        self._bottom_right.y = self._top_left.y + val

    @pos.setter
    def pos(self, point):
        size = self.size
        self._top_left.x = point.x
        self._top_left.y = point.y
        self._bottom_right = point + size

    @size.setter
    def size(self, point):
        self.width = point.x
        self.height = point.y

    @center.setter
    def center(self, point):
        size = self.size
        self._top_left.x = point.x - size.x / 2
        self._top_left.y = point.y - size.y / 2
        self._bottom_right = self._top_left + size

    @x1.setter
    def x1(self, val):
        self._top_left.x = val

    @y1.setter
    def y1(self, val):
        self._top_left.y = val

    @x2.setter
    def x2(self, val):
        self._bottom_right.x = val

    @y2.setter
    def y2(self, val):
        self._bottom_right.y = val

    def __eq__(self, other):
        return self._top_left == other._top_left and self._bottom_right == other._bottom_right

    def __ne__(self, other):
        return not (self == other)

    def copy(self):
        return Rect(x1=self.x1, y1=self.y1, x2=self.x2, y2=self.y2)

    def __repr__(self):
        return "%s(%ix%i%+d%+i)" % (
            self.__class__.__name__,
            self.width,
            self.height,
            self.x,
            self.y
        )



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from math import sqrt, floor


class IndexedColor(object):
    """
    Color within a palette
    """
    def __init__(self, color, palette):
        if type(color) is str:
            self.index = palette.find_index(color)
        else:
            self.index = color

        self.palette = palette

    @property
    def rgb_tuple(self):
        """
        Returns a tuple with the rgb components
        """
        return self.palette.rgb_tuple(self.index)

    @property
    def rgb(self):
        """
        Returns a RgbColor object corresponding to this color
        """
        return self.palette.rgb(self.index)

    @property
    def name(self):
        """
        Color name as from the palette
        """
        return self.palette.name(self.index)

    def __eq__(self, oth):
        """
        Compares the rgb tuple of two colors to check whether they are the same
        """
        return self.rgb_tuple == RgbColor._color_tuple(oth)

    def __ne__(self, oth):
        """
        Compares the rgb tuple of two colors to check whether they are the same
        """
        return not (self == oth)

    def __hash__(self):
        return hash((IndexedColor, self.index, self.palette))


class RgbColor(object):
    """
    24 Bit RGB color
    """
    def __init__(self, r, g, b, name=None):
        self.r = r
        self.g = g
        self.b = b
        self._name = name

    @property
    def rgb_tuple(self):
        """
        Returns a tuple with the rgb components
        """
        return (self.r, self.g, self.b)

    @property
    def rgb(self):
        """
        Returns a RgbColor object corresponding to this color
        """
        return self

    @property
    def name(self):
        """
        Color name (if set) otherwise the hex string
        """
        return self._name if self._name is not None else self.hex()

    @name.setter
    def name(self, name):
        """
        Sets the color name to a specific string
        """
        self._name = name

    @name.deleter
    def name(self):
        """
        Resets the name to the default (which is the same as hex)
        """
        self._name = None

    def __eq__(self, oth):
        """
        Compares the rgb tuple of two colors to check whether they are the same
        """
        return self.rgb_tuple == RgbColor._color_tuple(oth)

    def __ne__(self, oth):

        """
        Compares the rgb tuple of two colors to check whether they are the same
        """
        return not (self == oth)

    def copy(self):
        """
        Returns a deep copy of this object
        """
        return RgbColor(self.r, self.g, self.b, self.name)

    def hex(self):
        """
        Hex string for this color
        """
        return '#%02x%02x%02x' % self.rgb_tuple

    @staticmethod
    def _color_tuple(obj):
        """
        Returns a rgb tuple for comaring color-like objects
        """
        if obj is None:
            return tuple()
        if type(obj) is tuple and len(obj) == 3:
            return obj
        return obj.rgb_tuple

    def __hash__(self):
        return hash((RgbColor, self.rgb_tuple))

    def rgb_float(self):
        """
        Returns normalized RGB components
        """
        return (
            self.r / 255.0,
            self.g / 255.0,
            self.b / 255.0,
        )

    def _xyz(self):
        """
        Converts the color into XYZ coordinates
        """
        redf, greenf, bluef = self.rgb_float()

        def conv(v):
            return ((v + 0.055) / 1.055) ** 2.4 if v > 0.04045 else v / 12.92
        redf = conv(redf) * 100
        greenf = conv(greenf) * 100
        bluef = conv(bluef) * 100

        X = redf * 0.4124 + greenf * 0.3576 + bluef * 0.1805
        Y = redf * 0.2126 + greenf * 0.7152 + bluef * 0.0722
        Z = redf * 0.0193 + greenf * 0.1192 + bluef * 0.9505

        return (X, Y, Z)

    def lab(self):
        """
        Converts the color into L*a*b* coordinates
        """
        X, Y, Z = self._xyz()

        ref_X =  95.047
        ref_Y = 100.000
        ref_Z = 108.883

        x = X / ref_X
        y = Y / ref_Y
        z = Z / ref_Z

        def conv(v):
            return v ** (1.0 / 3) if v > 0.008856 else (7.787 * v) + (16.0 / 116)
        x = conv(x)
        y = conv(y)
        z = conv(z)

        L = (116 * y) - 16
        a = 500 * (x - y)
        b = 200 * (y - z)

        def to_int(v):
            return int(round(v))
        return (to_int(L), to_int(a), to_int(b))

    def distance(self, other):
        """
        Evaluates the Delta-E distance between two RGB colors
        """
        return sqrt(sum(map(lambda a: (a[0] - a[1]) ** 2, zip(self.lab(), other.lab()))))

    def to_indexed(self, palette):
        """
        Returns the monst similar color from the palette
        """
        if not palette:
            raise ValueError("Cannot convert to an indexed color with an empty palette")

        min_distance = self.distance(palette.rgb(0))
        min_index = 0
        for index, color in enumerate(palette.colors):
            distance = self.distance(RgbColor(*color))
            if distance < min_distance:
                min_distance = distance
                min_index = index

        return IndexedColor(min_index, min_distance)

    @classmethod
    def from_hsv(cls, h, s, v):
        """
        Coverts from HSV components
        Each component shall be in [0, 1]
        """
        if h < 0:
            h = 0
        elif h > 1:
            h %= 1
        s = min(max(s, 0), 1)
        v = min(max(v, 0), 1)

        h *= 6
        c = v * s
        m = v - c

        h1 = floor(h)
        f = h - h1
        n = v - c * f
        k = v - c * (1 - f)

        v = int(round(v * 255))
        m = int(round(m * 255))
        n = int(round(n * 255))
        k = int(round(k * 255))

        if h1 == 0:
            return cls(v, k, m)
        elif h1 == 1:
            return cls(n, v, m)
        elif h1 == 2:
            return cls(m, v, k)
        elif h1 == 3:
            return cls(m, n, v)
        elif h1 == 4:
            return cls(k, m, v)
        elif h1 == 5:
            return cls(v, m, n)
        else: # h1 == 6
            return cls(v, k, m)

    def hsv(self):
        """
        Returns a tuple with normalized HSV components
        """
        r, g, b = self.rgb_float()
        if r == g == b:
            cmax = r
            h = 0
            delta = 0
        else:
            cmax = max(r, g, b)
            cmin = min(r, g, b)
            delta = cmax - cmin

            if cmax == r:
                h = ((g - b) / delta) % 6
            elif cmax == g:
                h = (b - r) / delta + 2
            else: # cmax == b
                h = (r - g) / delta + 4

        h /= 6
        s = delta / cmax if cmax > 0 else 0
        v = cmax

        return (h, s, v)


class UnchangedColorType(object):
    """
    Dummy class for the UnchangedColor constant
    """
    pass


UnchangedColor = UnchangedColorType()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import curses
from contextlib import contextmanager


@contextmanager
def curses_context():
    stdscr = curses.initscr()
    curses.noecho()
    curses.cbreak()
    curses.nonl()
    stdscr.keypad(1)
    curses.start_color()
    curses.use_default_colors()

    curses.init_pair(1, curses.COLOR_BLACK,     -1)
    curses.init_pair(2, curses.COLOR_RED,       -1)
    curses.init_pair(3, curses.COLOR_GREEN,     -1)
    curses.init_pair(4, curses.COLOR_YELLOW,    -1)
    curses.init_pair(5, curses.COLOR_BLUE,      -1)
    curses.init_pair(6, curses.COLOR_MAGENTA,   -1)
    curses.init_pair(7, curses.COLOR_CYAN,      -1)
    curses.init_pair(8, curses.COLOR_WHITE,     -1)

    try:
        yield stdscr
    finally:
        stdscr.keypad(0)
        curses.nl()
        curses.nocbreak()
        curses.echo()
        curses.endwin()


def color_to_curses(color):
    if color is None:
        return curses.A_NORMAL | curses.color_pair(0)
    flags = curses.color_pair((color.index & 7) + 1)
    if color.index & 8:
        flags |= curses.A_BOLD
    return flags
